# frozen_string_literal: true

require 'ice_cube'

# TODO: Take advantage of the ice_cube gem
#       ::IceCube::Rule.weekly.day(:friday).day_of_month(13).month_of_year(:october).to_yaml ?

module ResourcePlanner
  class Event < ApplicationRecord
    #attr_accessible :parent_event, :parent_event_id
    #attr_accessible :name, :description, :start_at, :end_at, :all_day
    #attr_accessible :ical_expression

    # data required for creating iCal definition
    attr_accessor :dtstart
    attr_accessor :dtend
    attr_accessor :duration
    attr_accessor :exdate
    attr_accessor :rdate
    attr_accessor :rrule
    attr_accessor :timezone

    # Recurrence rules
    attr_accessor :until
    attr_accessor :count
    attr_accessor :interval
    attr_accessor :bysecond
    attr_accessor :byminute
    attr_accessor :byhour
    attr_accessor :byday
    attr_accessor :bymonthday
    attr_accessor :byyearday
    attr_accessor :byweekno
    attr_accessor :bymonth
    attr_accessor :bysetpos
    attr_accessor :wkst
    attr_accessor :freq

    #--------------------------------------------------------------------------

    # Virtual attributes
    # ------------------
    attr_accessor :start_at_tmp
    attr_accessor :end_at_tmp
    attr_accessor :all_day_tmp

    def start_at=(value)
      if event_units.empty?
        @start_at_tmp = value
      else
        event_unit = event_units.order(:start_at).first
        event_unit = event_units.first if event_unit.nil?
        event_unit.start_at = value
      end
    end

    def start_at
      if event_units.empty?
        @start_at_tmp
      else
        event_unit = event_units.order(:start_at).first
        event_unit = event_units.first if event_unit.nil?
        event_unit.start_at
      end
    end

    def end_at=(value)
      if event_units.empty?
        @end_at_tmp = value
      else
        event_unit = event_units.order(:end_at).last
        event_unit = event_units.first if event_unit.nil?
        event_unit.end_at = value
      end
    end

    def end_at
      if event_units.empty?
        @end_at_tmp
      else
        event_unit = event_units.order(:end_at).last
        event_unit = event_units.first if event_unit.nil?
        event_unit.end_at
      end
    end

    def all_day=(value)
      if event_units.size == 1
        event_units.first.all_day = value
      else
        @all_day_tmp = value
      end
    end

    def all_day
      if event_units.size == 1
        event_units.first.all_day
      else
        @all_day_tmp
      end
    end

    #--------------------------------------------------------------------------

    #enum :event_type => [:event, :availability, :restriction]

    #has_event_calendar

    belongs_to :event_owner, polymorphic: true, optional: true

    belongs_to :parent_event, class_name: '::ResourcePlanner::Event', dependent: :destroy, optional: true

    belongs_to :event_type, class_name: '::ResourcePlanner::EventType'

    has_many :event_units, class_name: '::ResourcePlanner::EventUnit', foreign_key: 'event_id', autosave: true, dependent: :destroy
    accepts_nested_attributes_for :event_units, allow_destroy: true

    has_many :recurring_event_exceptions, class_name: '::ResourcePlanner::Event', foreign_key: 'parent_event_id', dependent: :destroy
    accepts_nested_attributes_for :recurring_event_exceptions, allow_destroy: true

    #--------------------------------------------------------------------------

    #---------- Validations

    #validates :event_units, :presence => true
    #validates :event_units, :size => { :minimum => 1 }
    #validates :event_units, :length => { :minimum => 1 }

    #---------- Scopes

    scope :events, -> { where(:event_type => ::ResourcePlanner::EventType.event) }
    scope :resource_availabilities, -> { where(:event_type => ::ResourcePlanner::EventType.availability) }
    scope :resource_restrictions, -> { where(:event_type => ::ResourcePlanner::EventType.restriction) }

    #---------- Callbacks

    after_initialize  :ensure_at_least_one_event_unit_exists
    before_validation :associate_and_update_event_units
    before_validation :set_ical_expression

    #--------------------------------------------------------------------------

    # TBD: Remove this when EventCalendar is fixed
    # Get the events overlapping the given start and end dates
    # def self.events_for_date_range(start_d, end_d, find_options = {})
    #   self.where(find_options)
    #   .where("(? <= #{self.quoted_table_name}.#{self.end_at_field}) AND (#{self.quoted_table_name}.#{self.start_at_field}< ?)", start_d.to_time.utc, end_d.to_time.utc)
    #   .order("#{self.quoted_table_name}.#{self.start_at_field} ASC")
    # end

    #--------------------------------------------------------------------------

    def to_s
      "#{name} (#{start_at}-#{end_at})"
    end

    def full_details
      ::IceCube::Rule.from_yaml(ical_expression)
    end

    def all_day?
      all_day
    end

    # are there any active conflicts
    def has_conflicts?
      conflicts_count > 0
    end

    # return count of active conflicts
    def conflicts_count
      conflicts.length
    end

    # return active conflicts
    def conflicts
      active_conflicts = []

      event_units.each do |event_unit|
        event_unit.resources.each do |event_unit_resource|
          # quantity consumed by this event unit
          event_unit_subscription = event_unit.subscriptions.where(resource: event_unit_resource).first
          total_resource_consumed = event_unit_subscription&.quantity ? event_unit_subscription.quantity : 0

          coincident_event_units = event_unit_resource.event_units
                                                      .where.not(event_id: id)
                                                      .where(start_at: event_unit.start_at..event_unit.end_at)
                                                      .where(end_at: event_unit.start_at..event_unit.end_at)

          # coincident_event_units = ::ResourcePlanner::EventUnit
          #                          .where.not(event_id: id)
          #                          .where(start_at: event_unit.start_at..event_unit.end_at)
          #                          .where(end_at: event_unit.start_at..event_unit.end_at)

          potential_conflicts = [event_unit]

          coincident_event_units.each do |coincident_event_unit|
            coincident_event_unit.subscriptions.each do |coincident_event_unit_subscription|
              potential_conflicts.push(coincident_event_unit)
              total_resource_consumed += coincident_event_unit_subscription.quantity if coincident_event_unit_subscription.quantity
            end
          end

          active_conflicts.push(potential_conflicts) if event_unit_resource.quantity && total_resource_consumed > event_unit_resource.quantity
        end
      end

      active_conflicts
    end

    #--------------------------------------------------------------------------

    private

    def to_ical_expression(count = 1)
      case repeating_name
      when 'secondly'
        ::IceCube::Rule.secondly(count).to_yaml
      when 'minutely'
        ::IceCube::Rule.minutely(count).to_yaml
      when 'hourly'
        ::IceCube::Rule.hourly(count).to_yaml
      when 'daily'
        ::IceCube::Rule.daily(count).to_yaml
      when 'weekly'
        ::IceCube::Rule.weekly(count).to_yaml
      when 'monthly'
        ::IceCube::Rule.monthly(count).to_yaml
      when 'annually'
        ::IceCube::Rule.yearly(count).to_yaml
      else
        ''
      end
    end

    #---------- Callback handlers

    # Ensure that there is at least one EventUnit
    def ensure_at_least_one_event_unit_exists
      if new_record? && event_units.empty?
        event_units.build
        event_units.first.attributes = {
          event: self
        }
        # event_units.build(
        #   start_at: start_at_tmp,
        #   end_at: end_at_tmp,
        #   all_day: all_day_tmp
        # )
        # could be: event_units.build([ {}, {}, {}, ... ])
      end
    end

    def associate_and_update_event_units
      event_units.each { |event_unit| event_unit.event = self }

      if event_units.length == 1 && !(start_at_tmp.nil? || end_at_tmp.nil?)
        event_units.first.attributes = {
          start_at: start_at_tmp,
          end_at: end_at_tmp,
          all_day: all_day_tmp
        }
      end
    end

    def set_ical_expression
      self.ical_expression = ::IceCube::Rule.yearly.day_of_month(23).day(:friday).month_of_year(:october).to_yaml
    end
  end
end
